<?php

namespace FiloBlu\ProductUrlTools\Console\Command;

use Exception;
use Magento\Catalog\Model\Product\Attribute\Source\Status;
use Magento\Catalog\Model\Product\Visibility;
use Magento\Catalog\Model\ResourceModel\Product\Action;
use Magento\Catalog\Model\ResourceModel\Product\ActionFactory;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\Framework\App\ResourceConnection;
use Psr\Log\LoggerInterface;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Helper\ProgressBar;

/**
 * Class RegenerateProductUrlKeyCommand
 * @package FiloBlu\ProductUrlTools\Console\Command
 */
class RegenerateProductUrlKeyCommand extends Command
{
    /**
     * @var CollectionFactory
     */
    protected $collectionFactory;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * @var Action
     */
    protected $action;

    /**
     * @var ResourceConnection
     */
    protected $resource;

    /**
     * @var Status
     */
    protected $productStatus;

    /**
     * @var Visibility
     */
    protected $productVisibility;

    /**
     * @var ActionFactory
     */
    protected $actionFactory;

    /**
     * @var ExecutorGroup
     */
    protected $executorGroup;

    /**
     * RegenerateProductUrlKeyCommand constructor.
     * @param CollectionFactory $collectionFactory
     * @param LoggerInterface $logger
     * @param Status $productStatus
     * @param Visibility $productVisibility
     * @param ResourceConnection $resource
     * @param ActionFactory $actionFactory
     * @param ExecutorGroup $executorGroup
     */
    public function __construct(
        CollectionFactory $collectionFactory,
        LoggerInterface $logger,
        Status $productStatus,
        Visibility $productVisibility,
        ResourceConnection $resource,
        ActionFactory $actionFactory,
        ExecutorGroup $executorGroup
    )
    {
        $this->collectionFactory = $collectionFactory;
        $this->logger = $logger;
        $this->productVisibility = $productVisibility;
        $this->productStatus = $productStatus;
        $this->actionFactory = $actionFactory;
        $this->resource = $resource;
        $this->executorGroup = $executorGroup;

        parent::__construct();
    }


    /**
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int|void|null
     */
    public function execute(InputInterface $input, OutputInterface $output)
    {
        $storesList = $this->getStoresToProcess($input->getOption('store'));
        $productIds = $input->getArgument('pids');
        $allVisible = $input->getOption('all_visible');
        $jobs = $input->getOption('jobs');
        $this->action = $this->actionFactory->create();

        if ($jobs <= 1) {
            $section = null;
            /*
            // Progress bar not supported on all Magento versions
            ProgressBar::setFormatDefinition('stores', 'Now processing %current%/%max% %message% ... giovanotto/a');
            $storeProgressBar = new ProgressBar($output->section(), count($storesList));
            $storeProgressBar->setFormat('stores');
            $storeProgressBar->start();
            $section = $output->section();
            */

            foreach ($storesList as $storeId) {
                //$storeProgressBar->setMessage("Store {$storeId}");
                //$storeProgressBar->advance();
                $this->regenerateUrlKey($storeId, $productIds, $allVisible, $section);
                //$section->clear();
            }

            //$storeProgressBar->finish();
            $output->writeln('<info>✌  Apposc`t!</info>');
            return;
        }

        $jobsArguments = [];

        $i = 0;
        foreach ($storesList as $storeId) {
            $jobsArguments[$i++ % $jobs][] = $storeId;
        }

        foreach ($jobsArguments as $job => $argument) {
            $this->executorGroup->addPhpCommand(sprintf('%s %s %s %s', 'bin/magento', $this->getName(), '-s', implode(',', $argument)));
        }

        $this->executorGroup->executeAndWaitAll();
    }

    /**
     * @param $storesId comma separate store ids
     * @return array
     */
    protected function getStoresToProcess($storesId)
    {
        $allStores = $this->getAllStoreIds();

        if ($storesId === null) {
            return $allStores;
        }

        $stores = explode(',', $storesId);

        $output = [];
        foreach ($stores as $store) {
            $storeId = (int)$store;
            $output[$storeId] = $allStores[$storeId];
        }
        return $output;

    }

    /**
     * @return array
     */
    public function getAllStoreIds()
    {
        $result = [];

        $sql = $this->resource->getConnection()->select()
            ->from($this->resource->getTableName('store'), ['store_id', 'code'])
            ->order('store_id', 'ASC');

        foreach ($this->resource->getConnection()->fetchAll($sql) as $row) {
            if (isset($row['store_id']) && (int)$row['store_id'] > 0) {
                $result[(int)$row['store_id']] = $row['store_id'];
            }
        }

        return $result;
    }

    /**
     * @param $storeId
     * @param $allVisible
     * @param $productIds
     * @param OutputInterface $output
     */
    protected function regenerateUrlKey($storeId, $allVisible, $productIds, OutputInterface $output = null)
    {
        $collection = $this->collectionFactory->create();

        $collection->setStore($storeId);

        if (!empty($productIds)) {
            $collection->addIdFilter($productIds);
        }

        $collection->addAttributeToSelect(['name', 'url_key']);
        $collection->addStoreFilter($storeId)->setStoreId($storeId);

        if ($allVisible) {
            $collection->addAttributeToFilter('status', ['in' => $this->productStatus->getVisibleStatusIds()]);
            $collection->setVisibility($this->productVisibility->getVisibleInSiteIds());
        }

        //$progressBar = new ProgressBar($output, $collection->count());
        //$progressBar->start();

        foreach ($collection->getItems() as $product) {

            if (!$product->getName()) {
                $this->logger->info('name is empty for product : ' . $product->getId() . ' in store view ' . $storeId);
                continue;
            }
            $name = $product->getName();

            $urlKey = $product->formatUrlKey($name . '-' . $storeId . $product->getId());
            $attributesData = ['url_key' => $urlKey];

            try {
                $this->action->updateAttributes([$product->getId()], $attributesData, $storeId);
                $this->logger->info('processed urlkey for product with id : ' . $product->getId() . ' storeID ' . $storeId);
            } catch (Exception $e) {
                $output->writeln('<error>Duplicated url for ' . $product->getId() . ' ' . $e->getMessage() . ' </error>');
            }
            //$progressBar->advance();

        }

        //$progressBar->finish();
    }

    /**
     *
     */
    protected function configure()
    {
        $this->setName('tools:regenerateurlkey')
            ->setDescription('Regenerate urlkey for given products')
            ->addArgument(
                'pids',
                InputArgument::IS_ARRAY,
                'Products to regenerate'
            )
            ->addOption(
                'store', 's',
                InputOption::VALUE_REQUIRED,
                'Use the specific Store View'
            )
            ->addOption(
                'all_visible', 'a',
                InputOption::VALUE_OPTIONAL,
                'Use for all product visible'
            )->addOption('jobs', 'j',
                InputOption::VALUE_OPTIONAL,
                'Jobs', 1);
    }
}

